



git clone https://github.com/HypothesisWorks/hypothesis.git src
IF %ERRORLEVEL% NEQ 0 exit /B 1
cd src/hypothesis-python
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -m pytest  tests/cover
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -m pytest  tests/datetime
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -m pytest  tests/fakefactory
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -m pytest  tests/numpy
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -m pytest  tests/pandas
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -m pytest  tests/py3
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -m pytest  tests/pytest -p pytester --runpytest subprocess
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
